import itertools
import math
import re

import ninetoothed.naming as naming
from ninetoothed.language import call
from ninetoothed.symbol import Symbol


class Tensor:
    num_instances = 0

    def __init__(
        self,
        ndim=None,
        shape=None,
        dtype=None,
        strides=None,
        other=None,
        constexpr_shape=None,
        name=None,
        source=None,
        source_dims=None,
        target=None,
        target_dims=None,
    ):
        self.dtype = dtype

        if name is not None:
            self.name = name
        else:
            self.name = naming.auto_generate(f"tensor_{type(self).num_instances}")

        if ndim is not None:
            self.shape = (
                Symbol(self.size_string(i), constexpr=constexpr_shape)
                for i in range(ndim)
            )
            self.strides = (Symbol(self.stride_string(i)) for i in range(ndim))
        else:
            self.shape = shape

            if strides is not None:
                self.strides = strides
            else:
                self.strides = self._calculate_default_strides(shape)

        self.other = other

        if source is not None:
            self.source = source
        else:
            self.source = self

        if source_dims is not None:
            self.source_dims = source_dims
        else:
            self.source_dims = (dim for dim in range(self.source.ndim))

        if target is not None:
            self.target = target
        else:
            self.target = self

        if target_dims is not None:
            self.target_dims = target_dims
        else:
            self.target_dims = (dim for dim in range(self.target.ndim))

        type(self).num_instances += 1

    def tile(self, tile_shape, strides=None, dilation=None):
        if strides is None:
            strides = [-1 for _ in tile_shape]

        if dilation is None:
            dilation = [1 for _ in tile_shape]

        outer_shape = []
        outer_strides = []
        inner_shape = []
        inner_strides = []

        for self_size, self_stride, tile_size, stride, spacing in zip(
            self.shape, self.strides, tile_shape, strides, dilation
        ):
            if tile_size == -1:
                tile_size = self_size

            if stride == -1:
                stride = tile_size

            new_size = (
                call("cdiv", self_size - spacing * (tile_size - 1) - 1, stride) + 1
                if stride != 0
                else -1
            )
            outer_shape.append(new_size)

            new_stride = self_stride * stride // spacing
            outer_strides.append(new_stride)

            inner_shape.append(tile_size)
            next_stride = self_stride * spacing
            inner_strides.append(next_stride)

        return type(self)(
            shape=outer_shape,
            dtype=type(self)(
                shape=inner_shape,
                dtype=self.dtype,
                strides=inner_strides,
                source=self.source,
                source_dims=self.source_dims,
            ),
            strides=outer_strides,
            source=self.source,
            source_dims=self.source_dims,
        )

    def expand(self, shape):
        # TODO: Add error handling.
        return type(self)(
            shape=[
                new_size if new_size != -1 else size
                for size, new_size in zip(self.shape, shape)
            ],
            dtype=self.dtype,
            strides=[
                stride if new_size == -1 else 0
                for new_size, stride in zip(shape, self.strides)
            ],
            source=self.source,
            source_dims=self.source_dims,
        )

    def squeeze(self, dim):
        if not isinstance(dim, tuple):
            dim = (dim,)

        # TODO: Add error handling.
        return type(self)(
            shape=[size for i, size in enumerate(self.shape) if i not in dim],
            dtype=self.dtype,
            strides=[stride for i, stride in enumerate(self.strides) if i not in dim],
            source=self.source,
            source_dims=[
                source_dim
                for i, source_dim in enumerate(self.source_dims)
                if i not in dim
            ],
        )

    def permute(self, dims):
        # TODO: Add error handling.
        new_shape = [None for _ in range(self.ndim)]
        new_strides = [None for _ in range(self.ndim)]
        new_source_dims = [None for _ in range(self.ndim)]

        for original_dim, permuted_dim in enumerate(dims):
            new_shape[original_dim] = self.shape[permuted_dim]
            new_strides[original_dim] = self.strides[permuted_dim]
            new_source_dims[original_dim] = self.source_dims[permuted_dim]

        return type(self)(
            shape=new_shape,
            dtype=self.dtype,
            strides=new_strides,
            source=self.source,
            source_dims=new_source_dims,
        )

    def flatten(self, start_dim=None, end_dim=None):
        # TODO: Add error handling.
        if start_dim is None:
            start_dim = 0
        if end_dim is None:
            end_dim = self.ndim

        leading_sizes = self.shape[:start_dim]
        flattening_sizes = self.shape[start_dim:end_dim]
        trailing_sizes = self.shape[end_dim:]

        new_shape = leading_sizes + (math.prod(flattening_sizes),) + trailing_sizes

        leading_strides = self.strides[:start_dim]
        flattening_strides = self.strides[start_dim:end_dim]
        trailing_strides = self.strides[end_dim:]

        new_strides = leading_strides + (flattening_strides[-1],) + trailing_strides

        leading_source_dims = self.source_dims[:start_dim]
        flattening_source_dims = self.source_dims[start_dim:end_dim]
        trailing_source_dims = self.source_dims[end_dim:]

        new_source_dims = (
            leading_source_dims + (flattening_source_dims,) + trailing_source_dims
        )

        return type(self)(
            shape=new_shape,
            dtype=self.dtype,
            strides=new_strides,
            source=self.source,
            source_dims=new_source_dims,
        )

    def ravel(self):
        # TODO: Add error handling.
        new_shape = []
        new_strides = []

        curr = self

        while isinstance(curr, type(self)):
            new_shape.extend(curr.shape)
            new_strides.extend(curr.strides)

            curr = curr.dtype

        return type(self)(
            shape=new_shape,
            strides=new_strides,
            other=self.source.other,
            name=self.source.name,
        )

    def names(self):
        if self.ndim == 0:
            return {self.source.name}

        return (
            {self.source.pointer_string()}
            | {
                name
                for value in itertools.chain(self.shape, self.strides)
                if isinstance(value, Symbol)
                for name in value.names()
            }
            | (self.dtype.names() if isinstance(self.dtype, type(self)) else set())
            | (self.source.names() if self.source is not self else set())
        )

    def innermost(self):
        if not isinstance(self.dtype, type(self)):
            return self

        return self.dtype.innermost()

    def pointer_string(self):
        return f"{self.name}_pointer"

    def size_string(self, dim):
        return f"{self.name}_size_{dim}"

    def stride_string(self, dim):
        return f"{self.name}_stride_{dim}"

    def size(self, dim=None):
        if dim is None:
            return self.shape

        return self.shape[dim]

    def stride(self, dim=None):
        if dim is None:
            return self.strides

        return self.strides[dim]

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        self._shape = tuple(value)

    @property
    def strides(self):
        return self._strides

    @strides.setter
    def strides(self, value):
        self._strides = tuple(value)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def source_dims(self):
        return self._source_dims

    @source_dims.setter
    def source_dims(self, value):
        self._source_dims = tuple(value)

    @property
    def target_dims(self):
        return self._target_dims

    @target_dims.setter
    def target_dims(self, value):
        self._target_dims = tuple(value)

    @staticmethod
    def pointer_pattern():
        return re.compile(rf"({_identifier_pattern_raw_string()})_(pointer)")

    @staticmethod
    def size_pattern():
        return re.compile(rf"({_identifier_pattern_raw_string()})_(size)_(.+)")

    @staticmethod
    def stride_pattern():
        return re.compile(rf"({_identifier_pattern_raw_string()})_(stride)_(.+)")

    @staticmethod
    def _calculate_default_strides(shape):
        strides = [1]

        for size in reversed(shape[1:]):
            strides.append(size * strides[-1])

        return reversed(strides)


def _identifier_pattern_raw_string():
    return r"[a-zA-Z_][a-zA-Z0-9_]*"



import ast

from ninetoothed.symbol import Symbol

LANGUAGE = "ninetoothed.language"


def call(func, *args, **kwargs):
    return Symbol(
        ast.Call(
            func=attribute(func).node,
            args=[Symbol(arg).node for arg in args],
            keywords=[
                ast.keyword(arg=kwarg, value=Symbol(kwargs[kwarg]).node)
                for kwarg in kwargs
            ],
        )
    )


def attribute(attr):
    return Symbol(ast.parse(f"{LANGUAGE}.{attr}", mode="eval").body)



import re


def auto_generate(name):
    return f"ninetoothed_{name}"


def make_constexpr(name):
    return _add_prefix(name, _CONSTEXPR)


def make_meta(name):
    return _add_prefix(name, _META)


def make_next_power_of_2(name):
    return _add_prefix(name, _NEXT_POWER_OF_2)


def is_constexpr(name):
    return _CONSTEXPR in _find_prefixes(name) or is_meta(name)


def is_meta(name):
    return _META in _find_prefixes(name)


def is_next_power_of_2(name):
    return _NEXT_POWER_OF_2 in _find_prefixes(name)


def remove_prefixes(name):
    return _PREFIX_PATTERN.sub("", name)


_CONSTEXPR = "constexpr"

_META = "meta"

_NEXT_POWER_OF_2 = "next_power_of_2"

_PREFIX_PATTERN = re.compile(r"ninetoothed_((?!_).*?)_prefix_")


def _add_prefix(name, string):
    return f"{_make_prefix(string)}{name}"


def _make_prefix(string):
    return f"ninetoothed_{string}_prefix_"


def _find_prefixes(name):
    return set(_PREFIX_PATTERN.findall(name))



import torch

import ninetoothed
from ninetoothed import Symbol, Tensor
from tests.skippers import skip_if_cuda_not_available


def add(lhs, rhs):
    BLOCK_SIZE = Symbol("BLOCK_SIZE", meta=True)

    @ninetoothed.jit
    def add_kernel(
        lhs: Tensor(1).tile((BLOCK_SIZE,)),
        rhs: Tensor(1).tile((BLOCK_SIZE,)),
        output: Tensor(1).tile((BLOCK_SIZE,)),
    ):
        output = lhs + rhs  # noqa: F841

    output = torch.empty_like(lhs)

    add_kernel(lhs, rhs, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        size = 98432

        cls.lhs = torch.rand(size, device="cuda")
        cls.rhs = torch.rand(size, device="cuda")

    def test_fp32(self):
        lhs = type(self).lhs.to(torch.float32)
        rhs = type(self).rhs.to(torch.float32)

        assert torch.allclose(add(lhs, rhs), lhs + rhs)



import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Symbol, Tensor
from tests.skippers import skip_if_cuda_not_available


def softmax(input):
    BLOCK_SIZE = Symbol("BLOCK_SIZE", constexpr=True)

    @ninetoothed.jit
    def softmax_kernel(
        input_row: Tensor(2, other=float("-inf")).tile((1, BLOCK_SIZE)),
        output_row: Tensor(2).tile((1, BLOCK_SIZE)),
    ):
        row_minus_max = input_row - ntl.max(input_row)
        numerator = ntl.exp(row_minus_max)
        denominator = ntl.sum(numerator)
        output_row = numerator / denominator  # noqa: F841

    output = torch.empty_like(input)

    softmax_kernel(input, output, BLOCK_SIZE=input.shape[-1])

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        cls.input = torch.randn(1823, 781, device="cuda")

    def test_fp32(self):
        input = type(self).input.to(torch.float32)

        assert torch.allclose(softmax(input), torch.softmax(input, axis=-1))



import random

import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Symbol, Tensor
from tests.skippers import skip_if_cuda_not_available, skip_if_float8_e5m2_not_supported


def addmm(input, mat1, mat2, beta=1, alpha=1):
    BLOCK_SIZE_M = Symbol("BLOCK_SIZE_M", meta=True)
    BLOCK_SIZE_N = Symbol("BLOCK_SIZE_N", meta=True)
    BLOCK_SIZE_K = Symbol("BLOCK_SIZE_K", meta=True)

    input_tiled = Tensor(2).tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    output_tiled = Tensor(2).tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    mat1_tiled = (
        Tensor(2)
        .tile((BLOCK_SIZE_M, BLOCK_SIZE_K))
        .tile((1, -1))
        .expand((-1, output_tiled.shape[1]))
    )
    mat1_tiled.dtype = mat1_tiled.dtype.squeeze(0)

    mat2_tiled = (
        Tensor(2)
        .tile((BLOCK_SIZE_K, BLOCK_SIZE_N))
        .tile((-1, 1))
        .expand((output_tiled.shape[0], -1))
    )
    mat2_tiled.dtype = mat2_tiled.dtype.squeeze(1)

    @ninetoothed.jit
    def addmm_kernel(
        input: input_tiled,
        mat1: mat1_tiled,
        mat2: mat2_tiled,
        beta: Tensor(0),
        alpha: Tensor(0),
        output: output_tiled,
    ):
        accumulator = ntl.zeros(output.shape, dtype=ntl.float32)
        for k in range(mat1.shape[0]):
            accumulator += ntl.dot(mat1[k], mat2[k])
        output = beta * input + alpha * accumulator.to(ntl.float16)

    output = torch.empty(
        (mat1.shape[0], mat2.shape[1]), device=mat1.device, dtype=torch.float16
    )

    addmm_kernel(input, mat1, mat2, beta, alpha, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        shape = (512, 512)

        cls.input = torch.randn(shape, device="cuda")
        cls.mat1 = torch.randn(shape, device="cuda")
        cls.mat2 = torch.randn(shape, device="cuda")
        cls.beta = random.uniform(0, 1)
        cls.alpha = random.uniform(0, 1)

    def test_fp16(self):
        input = type(self).input.to(torch.float16)
        mat1 = type(self).mat1.to(torch.float16)
        mat2 = type(self).mat2.to(torch.float16)
        beta = type(self).beta
        alpha = type(self).alpha

        assert torch.allclose(
            addmm(input, mat1, mat2, beta=beta, alpha=alpha),
            torch.addmm(input, mat1, mat2, beta=beta, alpha=alpha),
            atol=0.075,
        )

    @skip_if_float8_e5m2_not_supported
    def test_fp8(self):
        input = type(self).input.to(torch.float8_e5m2)
        mat1 = type(self).mat1.to(torch.float8_e5m2)
        mat2 = type(self).mat2.T.to(torch.float8_e5m2)
        beta = type(self).beta
        alpha = type(self).alpha

        assert torch.allclose(
            addmm(input, mat1, mat2, beta=beta, alpha=alpha),
            torch.addmm(
                input.to(torch.float16),
                mat1.to(torch.float16),
                mat2.to(torch.float16),
                beta=beta,
                alpha=alpha,
            ),
            atol=0.125,
        )



import ninetoothed.naming as naming


def test_make_constexpr():
    assert naming.make_constexpr(_NAME) == f"ninetoothed_constexpr_prefix_{_NAME}"


def test_make_meta():
    assert naming.make_meta(_NAME) == f"ninetoothed_meta_prefix_{_NAME}"


def test_make_next_power_of_2():
    assert (
        naming.make_next_power_of_2(_NAME)
        == f"ninetoothed_next_power_of_2_prefix_{_NAME}"
    )
    assert (
        naming.make_next_power_of_2(naming.make_constexpr(_NAME))
        == f"ninetoothed_next_power_of_2_prefix_ninetoothed_constexpr_prefix_{_NAME}"
    )


def test_is_constexpr():
    assert naming.is_constexpr(naming.make_constexpr(_NAME))
    assert naming.is_constexpr(naming.make_meta(_NAME))


def test_is_meta():
    assert naming.is_meta(naming.make_meta(_NAME))


def test_is_next_power_of_2():
    assert naming.is_next_power_of_2(naming.make_next_power_of_2(_NAME))
    assert naming.is_next_power_of_2(
        naming.make_next_power_of_2(naming.make_constexpr(_NAME))
    )


def test_remove_prefixes():
    assert naming.remove_prefixes(naming.make_constexpr(_NAME)) == _NAME
    assert naming.remove_prefixes(naming.make_meta(_NAME)) == _NAME
    assert naming.remove_prefixes(naming.make_next_power_of_2(_NAME)) == _NAME
    assert (
        naming.remove_prefixes(
            naming.make_next_power_of_2(naming.make_constexpr(_NAME))
        )
        == _NAME
    )


_NAME = "ninetoothed_name"



import ast

import ninetoothed.naming as naming
from ninetoothed.tensor import Tensor


class Torchifier(ast.NodeTransformer):
    def visit_Name(self, node):
        self.generic_visit(node)

        source = node.id

        if naming.is_constexpr(source):
            return node

        def repl(match):
            return f"{match.group(1)}"

        source = Tensor.pointer_pattern().sub(repl, source)

        def repl(match):
            return f"{match.group(1)}.{match.group(2)}({match.group(3)})"

        source = Tensor.size_pattern().sub(repl, source)
        source = Tensor.stride_pattern().sub(repl, source)

        if source != node.id:
            return ast.parse(source, mode="eval").body

        return node

    def visit_Attribute(self, node):
        self.generic_visit(node)

        if (
            isinstance(node.value, ast.Name)
            and node.value.id == "ninetoothed"
            and node.attr == "language"
        ):
            return node.value

        return node



import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Symbol, Tensor
from tests.skippers import skip_if_cuda_not_available, skip_if_float8_e5m2_not_supported


def matmul(lhs, rhs):
    BLOCK_SIZE_M = Symbol("BLOCK_SIZE_M", meta=True)
    BLOCK_SIZE_N = Symbol("BLOCK_SIZE_N", meta=True)
    BLOCK_SIZE_K = Symbol("BLOCK_SIZE_K", meta=True)

    output_tiled = Tensor(2).tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    lhs_tiled = (
        Tensor(2)
        .tile((BLOCK_SIZE_M, BLOCK_SIZE_K))
        .tile((1, -1))
        .expand((-1, output_tiled.shape[1]))
    )
    lhs_tiled.dtype = lhs_tiled.dtype.squeeze(0)

    rhs_tiled = (
        Tensor(2)
        .tile((BLOCK_SIZE_K, BLOCK_SIZE_N))
        .tile((-1, 1))
        .expand((output_tiled.shape[0], -1))
    )
    rhs_tiled.dtype = rhs_tiled.dtype.squeeze(1)

    @ninetoothed.jit
    def matmul_kernel(lhs: lhs_tiled, rhs: rhs_tiled, output: output_tiled):
        accumulator = ntl.zeros(output.shape, dtype=ntl.float32)
        for k in range(lhs.shape[0]):
            accumulator += ntl.dot(lhs[k], rhs[k])
        output = accumulator.to(ntl.float16)

    output = torch.empty(
        (lhs.shape[0], rhs.shape[1]), device=lhs.device, dtype=torch.float16
    )

    matmul_kernel(lhs, rhs, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        shape = (512, 512)

        cls.lhs = torch.randn(shape, device="cuda")
        cls.rhs = torch.randn(shape, device="cuda")

    def test_fp16(self):
        lhs = type(self).lhs.to(torch.float16)
        rhs = type(self).rhs.to(torch.float16)

        assert torch.allclose(matmul(lhs, rhs), torch.matmul(lhs, rhs))

    @skip_if_float8_e5m2_not_supported
    def test_fp8(self):
        lhs = type(self).lhs.to(torch.float8_e5m2)
        rhs = type(self).rhs.T.to(torch.float8_e5m2)

        assert torch.allclose(
            matmul(lhs, rhs),
            torch.matmul(lhs.to(torch.float16), rhs.to(torch.float16)),
            atol=0.125,
        )



import pytest
import torch


def skip_if_cuda_not_available(func):
    return pytest.mark.skipif(
        not torch.cuda.is_available(),
        reason="CUDA not available",
    )(func)


def skip_if_float8_e5m2_not_supported(func):
    return pytest.mark.skipif(
        not hasattr(torch, "float8_e5m2"),
        reason="`float8_e5m2` not supported",
    )(func)



import ast
import collections
import copy
import functools
import importlib.util
import inspect
import itertools
import math
import subprocess
import sys
import tempfile

import triton

import ninetoothed.naming as naming
from ninetoothed.language import attribute, call
from ninetoothed.symbol import Symbol
from ninetoothed.tensor import Tensor
from ninetoothed.torchifier import Torchifier


def make(arrangement, application, tensors):
    params = inspect.signature(application).parameters
    types = arrangement(*tensors)
    annotations = {param: type for param, type in zip(params, types)}
    application.__annotations__ = annotations

    return jit(application)


def jit(_func=None, *, _prettify=False):
    def wrapper(func):
        return JIT(func, _prettify=_prettify)()

    if _func is None:
        return wrapper

    return wrapper(_func)


class JIT:
    def __init__(self, func, _prettify=False):
        self.func = func

        self._prettify = _prettify

    def __call__(self):
        tree = self._get_tree()

        CodeGenerator(inspect.get_annotations(self.func)).visit(tree)
        Tritonizer().visit(tree)
        _BinOpSimplifier().visit(tree)
        ast.fix_missing_locations(tree)

        if self._prettify:
            name_collector = _SimplifiedNameCollector()
            name_collector.visit(tree)

        unparsed = ast.unparse(tree).replace("None:", ":").replace(":None", ":")
        dependencies = self._find_dependencies()
        source = "\n\n".join((unparsed, dependencies)).strip()

        if self._prettify:
            for original, simplified in name_collector.simplified_names.items():
                if simplified not in name_collector.simplified_names:
                    source = source.replace(original, simplified)

            source = subprocess.check_output(
                ["ruff", "format", "-"], input=source, encoding="utf-8"
            )

        with tempfile.NamedTemporaryFile(delete=False, suffix=".py") as temp_file:
            temp_file.write(source.encode("utf-8"))
            temp_file_name = temp_file.name

        module = type(self)._import_from_path(temp_file_name, temp_file_name)
        module_vars = vars(module)

        handle = _Handle(
            module_vars[self.func.__name__],
            module_vars[f"launch_{self.func.__name__}"],
            source,
        )

        return handle

    def _get_tree(self):
        module = ast.parse(inspect.getsource(inspect.getmodule(self.func)))

        _AliasRestorer().visit(module)
        collector = _ImportCollector()
        collector.visit(module)
        finder = _FunctionDefFinder(self.func.__name__)
        finder.visit(module)

        return ast.Module(body=collector.imports + [finder.result], type_ignores=[])

    def _find_dependencies(self):
        dependencies = set()

        for obj in self.func.__globals__.values():
            if isinstance(obj, triton.runtime.JITFunction):
                dependencies.add(obj.src)

        return "\n".join(f"@triton.jit\n{dependency}" for dependency in dependencies)

    @staticmethod
    def _import_from_path(module_name, file_path):
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        return module


class CodeGenerator(ast.NodeTransformer):
    def __init__(self, context):
        super().__init__()

        self._context = context

        self._args = list(self._context.values())

        self._POWER_OF_TWOS = tuple(2**n for n in range(5, 11))

        self._MIN_PRODUCT = 2**10

        self._MAX_PRODUCT = 2**20

    def visit_Module(self, node):
        self.generic_visit(node)

        node.body.append(self._launch)

        return node

    def visit_FunctionDef(self, node):
        self._func_def = node

        self._invariants = {}

        self.generic_visit(node)

        for target, value in reversed(self._invariants.items()):
            node.body.insert(0, ast.Assign(targets=[target.node], value=value.node))

        return node

    def visit_arguments(self, node):
        self.generic_visit(node)

        names_of_args = [arg.names() - {"ninetoothed"} for arg in self._args]
        names = functools.reduce(lambda x, y: x | y, names_of_args)
        meta_names = {name for name in names if naming.is_meta(name)}
        non_meta_names = {name for name in names if name not in meta_names}
        non_meta_names |= {
            naming.make_next_power_of_2(name)
            for name in non_meta_names
            if naming.is_constexpr(name)
        }

        node.args = [
            ast.arg(arg=name)
            if not naming.is_constexpr(name)
            else ast.arg(arg=name, annotation=attribute("constexpr").node)
            for name in non_meta_names
        ] + [
            ast.arg(arg=name, annotation=attribute("constexpr").node)
            for name in meta_names
        ]

        autotune = self._generate_autotune(non_meta_names, meta_names)
        self._func_def.decorator_list = [autotune, Symbol("triton.jit").node]

        self._launch = self._generate_launch(non_meta_names, meta_names)

        return node

    def visit_Subscript(self, node):
        if self._in_context(node.value) and isinstance(node.ctx, ast.Load):
            value = self._context[node.value.id]

            if isinstance(value, Tensor):
                return self._generate_load(
                    value,
                    indices=node.slice.elts
                    if isinstance(node.slice, ast.Tuple)
                    else (node.slice,),
                )

        self.generic_visit(node)

        return node

    def visit_Attribute(self, node):
        if self._in_context(node.value):
            value = self._context[node.value.id]

            if isinstance(value, Tensor):
                inner = value.dtype

                return Symbol(getattr(inner, node.attr)).node

        self.generic_visit(node)

        return node

    def visit_Name(self, node):
        self.generic_visit(node)

        if self._in_context(node) and isinstance(node.ctx, ast.Load):
            return self._generate_load(self._context[node.id])

        return node

    def visit_Assign(self, node):
        if len(node.targets) == 1:
            target = node.targets[0]

            if self._in_context(target):
                self.generic_visit(node)

                return ast.Expr(
                    self._generate_store(self._context[target.id], node.value)
                )
            elif (
                isinstance(target, ast.Subscript)
                and self._in_context(target.value)
                and isinstance(target.ctx, ast.Store)
            ):
                value = self._context[target.value.id]

                if isinstance(value, Tensor):
                    self.generic_visit(node)

                    return ast.Expr(
                        self._generate_store(
                            value,
                            node.value,
                            indices=target.slice.elts
                            if isinstance(target.slice, ast.Tuple)
                            else (target.slice,),
                        )
                    )

        self.generic_visit(node)

        return node

    _NAME_FOR_PID = Symbol("ninetoothed_pid")

    def _in_context(self, node):
        return isinstance(node, ast.Name) and node.id in self._context

    def _generate_autotune(self, params, meta):
        device = triton.runtime.driver.active.get_current_device()
        properties = triton.runtime.driver.active.utils.get_device_properties(device)
        max_shared_mem = properties["max_shared_mem"]

        num_warps = 8
        num_stages = max_shared_mem // 2**15

        configs = [
            ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="ninetoothed", ctx=ast.Load()),
                    attr="Config",
                    ctx=ast.Load(),
                ),
                args=[
                    ast.Dict(
                        keys=[ast.Constant(value=param) for param in meta],
                        values=[ast.Constant(value=value) for value in values],
                    )
                ],
                keywords=[
                    ast.keyword(arg="num_warps", value=ast.Constant(value=num_warps)),
                    ast.keyword(arg="num_stages", value=ast.Constant(value=num_stages)),
                ],
            )
            for values in itertools.product(self._POWER_OF_TWOS, repeat=len(meta))
            if self._MIN_PRODUCT <= math.prod(values) <= self._MAX_PRODUCT
        ]

        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="ninetoothed", ctx=ast.Load()),
                attr="autotune",
                ctx=ast.Load(),
            ),
            args=[],
            keywords=[
                ast.keyword(
                    arg="configs",
                    value=ast.List(
                        elts=configs,
                        ctx=ast.Load(),
                    ),
                ),
                ast.keyword(
                    arg="key",
                    value=ast.List(
                        elts=[
                            ast.Constant(value=param)
                            for param in params
                            if not Tensor.pointer_pattern().fullmatch(param)
                        ],
                        ctx=ast.Load(),
                    ),
                ),
            ],
        )

    def _generate_launch(self, params, meta):
        non_next_power_of_2_constexpr_params = [
            param
            for param in params
            if naming.is_constexpr(param) and not naming.is_next_power_of_2(param)
        ]
        non_next_power_of_2_constexpr_params_without_prefixes = [
            naming.remove_prefixes(param)
            for param in non_next_power_of_2_constexpr_params
        ]
        next_power_of_2_params = [
            param for param in params if naming.is_next_power_of_2(param)
        ]
        next_power_of_2_params_without_prefixes = [
            naming.remove_prefixes(param) for param in next_power_of_2_params
        ]

        launch = ast.FunctionDef(
            name=f"launch_{self._func_def.name}",
            args=ast.arguments(
                posonlyargs=[],
                args=[ast.arg(arg=arg.source.name) for arg in self._args]
                + [
                    ast.arg(arg=param)
                    for param in non_next_power_of_2_constexpr_params_without_prefixes
                    if not Tensor.size_pattern().fullmatch(param)
                ],
                kwonlyargs=[],
                defaults=[],
            ),
            body=[
                ast.Assign(
                    targets=[ast.Name(id=param, ctx=ast.Store())],
                    value=ast.Name(id=param_without_prefixes, ctx=ast.Load()),
                )
                for param, param_without_prefixes in zip(
                    non_next_power_of_2_constexpr_params,
                    non_next_power_of_2_constexpr_params_without_prefixes,
                )
            ]
            + [
                ast.Assign(
                    targets=[ast.Name(id=param, ctx=ast.Store())],
                    value=Symbol(
                        f"triton.next_power_of_2({param_without_prefixes})"
                    ).node,
                )
                for param, param_without_prefixes in zip(
                    next_power_of_2_params,
                    next_power_of_2_params_without_prefixes,
                )
            ]
            + [
                ast.Expr(
                    ast.Call(
                        func=ast.Subscript(
                            value=ast.Name(id=self._func_def.name, ctx=ast.Load()),
                            slice=self._generate_grid(),
                            ctx=ast.Load(),
                        ),
                        args=[ast.Name(id=param, ctx=ast.Load()) for param in params],
                        keywords=[],
                    )
                )
            ],
            decorator_list=[],
        )

        class MetaEncloser(ast.NodeTransformer):
            def __init__(self, meta):
                self._meta = meta

            def visit_Name(self, node):
                self.generic_visit(node)

                if node.id in self._meta:
                    return ast.Subscript(
                        value=ast.Name(id="meta", ctx=ast.Load()),
                        slice=ast.Constant(value=node.id),
                        ctx=ast.Load(),
                    )

                return node

        MetaEncloser(meta).visit(launch)

        Torchifier().visit(launch)

        return launch

    def _generate_grid(self):
        num_elements = functools.reduce(lambda x, y: x * y, self._args[0].shape)

        return ast.parse(f"lambda meta: ({num_elements},)", mode="eval").body

    def _generate_load(self, tensor, indices=()):
        if tensor.ndim == 0:
            return Symbol(tensor.source.name).node

        pointers, mask = self._generate_pointers_and_mask(tensor, indices)
        other = type(self)._generate_other(tensor)

        return call("load", pointers, mask=mask, other=other).node

    def _generate_store(self, tensor, value, indices=()):
        pointers, mask = self._generate_pointers_and_mask(tensor, indices)

        return call("store", pointers, value, mask=mask).node

    def _generate_pointers_and_mask(self, tensor, indices):
        invariant_target_dims = type(self)._find_invariant_target_dims(tensor)

        indices = self._complete_indices(tensor, indices)
        offsets = type(self)._generate_offsets(tensor, indices)

        for source_dim in range(tensor.source.ndim):
            for target_dim in range(tensor.target.ndim):
                if target_dim not in invariant_target_dims:
                    continue

                name = type(self)._name_for_offsets(tensor, source_dim, target_dim)
                self._invariants[name] = offsets[source_dim][target_dim]
                offsets[source_dim][target_dim] = name

        name_for_pointers = type(self)._name_for_pointers(tensor)
        self._invariants[name_for_pointers] = Symbol(tensor.source.pointer_string())

        for source_dim in range(tensor.source.ndim):
            for target_dim in range(tensor.target.ndim):
                if target_dim not in invariant_target_dims:
                    continue

                self._invariants[name_for_pointers] += (
                    offsets[source_dim][target_dim][
                        type(self)._generate_slices(tensor, target_dim)
                    ]
                    * tensor.source.strides[source_dim]
                )

        pointers = name_for_pointers + sum(
            offsets[source_dim][target_dim][
                type(self)._generate_slices(tensor, target_dim)
            ]
            * tensor.source.strides[source_dim]
            for source_dim in range(tensor.source.ndim)
            for target_dim in range(tensor.target.ndim)
            if target_dim not in invariant_target_dims
            and offsets[source_dim][target_dim] != 0
        )
        mask = functools.reduce(
            lambda x, y: x & y,
            (
                offsets[source_dim][target_dim][
                    type(self)._generate_slices(tensor, target_dim)
                ]
                < tensor.source.shape[source_dim]
                for source_dim in range(tensor.source.ndim)
                for target_dim in range(tensor.target.ndim)
                if offsets[source_dim][target_dim] != 0
            ),
        )

        return pointers, mask

    def _complete_indices(self, tensor, indices):
        indices = list(self._generate_pid_indices(tensor) + tuple(indices))

        for size in tensor.innermost().shape:
            if Symbol.is_name(size):
                name = size.node.id
                if not naming.is_meta(name):
                    size = naming.make_next_power_of_2(name)

            indices.append(call("arange", 0, size))

        return tuple(indices)

    def _generate_pid_indices(self, tensor):
        self._invariants[type(self)._NAME_FOR_PID] = call("program_id", 0)

        indices = list(
            type(self)._unravel_index(type(self)._NAME_FOR_PID, tensor.shape)
        )

        for dim, index in enumerate(indices):
            name = type(self)._name_for_index(tensor, dim)
            self._invariants[name] = index
            indices[dim] = name

        return tuple(indices)

    @staticmethod
    def _generate_other(tensor):
        other = tensor.source.other

        if isinstance(other, float) and not math.isfinite(other):
            return f"float('{other}')"

        return other

    @staticmethod
    def _generate_slices(tensor, dim):
        return tuple(
            slice(None) if target_dim == dim else None
            for target_dim in tensor.innermost().target_dims
        )

    @staticmethod
    def _generate_offsets(tensor, indices):
        offsets = collections.defaultdict(
            lambda: collections.defaultdict(lambda: Symbol(0))
        )

        curr = tensor
        start = 0

        while isinstance(curr, type(tensor)):
            stop = start + curr.ndim
            curr_indices = indices[start:stop]

            for index, stride, source_dim, target_dim in zip(
                curr_indices, curr.strides, curr.source_dims, curr.target_dims
            ):
                offsets[source_dim][target_dim] += index * stride

            start = stop
            curr = curr.dtype

        for source_dim in tuple(offsets):
            for target_dim in tuple(offsets[source_dim]):
                if not isinstance(source_dim, tuple):
                    continue

                unraveled = CodeGenerator._unravel_index(
                    offsets[source_dim][target_dim],
                    tuple(tensor.source.shape[dim] for dim in source_dim),
                )

                for offs, dim in zip(unraveled, source_dim):
                    offsets[dim][target_dim] = offs

        for source_dim in range(tensor.source.ndim):
            for target_dim in range(tensor.target.ndim):
                offsets[source_dim][target_dim] = copy.deepcopy(
                    offsets[source_dim][target_dim]
                )
                offsets[source_dim][target_dim].find_and_replace(
                    Symbol(tensor.source.strides[source_dim]), Symbol(1)
                )

        return offsets

    @staticmethod
    def _find_invariant_target_dims(tensor):
        invariant_target_dims = set()

        curr = tensor.dtype

        while isinstance(curr.dtype, Tensor):
            for target_dim in range(curr.target.ndim):
                if target_dim not in curr.target_dims:
                    invariant_target_dims.add(target_dim)

            curr = curr.dtype

        return invariant_target_dims

    @staticmethod
    def _name_for_pointers(tensor):
        return Symbol(f"{tensor.source.name}_pointers")

    @staticmethod
    def _name_for_offsets(tensor, source_dim, target_dim):
        return Symbol(f"{tensor.source.name}_offsets_{source_dim}_{target_dim}")

    @staticmethod
    def _name_for_index(tensor, dim):
        return Symbol(f"{tensor.source.name}_index_{dim}")

    @staticmethod
    def _unravel_index(index, shape):
        indices = []

        for stride in Tensor(shape=shape).strides:
            indices.append(index // stride)
            index %= stride

        return tuple(indices)


class Tritonizer(ast.NodeTransformer):
    def visit_Module(self, node):
        self.generic_visit(node)

        node.body.insert(0, ast.Import(names=[ast.alias(name="triton.language")]))
        node.body.insert(0, ast.Import(names=[ast.alias(name="triton")]))

        return node

    def visit_Name(self, node):
        self.generic_visit(node)

        if node.id == "ninetoothed" or "ninetoothed." in node.id:
            node.id = node.id.replace("ninetoothed", "triton")

        return node

    def visit_Call(self, node):
        self.generic_visit(node)

        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "triton"
            and node.func.attr == "jit"
        ):
            return ast.Attribute(
                value=ast.Name(id="triton", ctx=ast.Load()), attr="jit", ctx=ast.Load()
            )

        return node


class _BinOpSimplifier(ast.NodeTransformer):
    def visit_BinOp(self, node):
        self.generic_visit(node)

        if isinstance(node.op, ast.Mult):
            left = Symbol(node.left)
            right = Symbol(node.right)

            if left == 0 or right == 0:
                return Symbol(0).node

            if left == 1:
                return node.right

            if right == 1:
                return node.left

        return node


class _SimplifiedNameCollector(ast.NodeVisitor):
    def __init__(self):
        self.simplified_names = {}

    def visit_Name(self, node):
        self.generic_visit(node)

        self.simplified_names[node.id] = naming.remove_prefixes(node.id)


class _Handle:
    def __init__(self, kernel, launch, source):
        self._kernel = kernel
        self._launch = launch
        self._source = source

    def __call__(self, *args, **kwargs):
        return self._launch(*args, **kwargs)


class _AliasRestorer(ast.NodeTransformer):
    def __init__(self):
        super().__init__()

        self._aliases = {}
        self._redefined = set()

    def visit_Import(self, node):
        for alias in node.names:
            if alias.asname:
                self._aliases[alias.asname] = alias.name

        return node

    def visit_ImportFrom(self, node):
        for alias in node.names:
            full_name = f"{node.module}.{alias.name}"
            if alias.asname:
                self._aliases[alias.asname] = full_name

        return node

    def visit_Assign(self, node):
        for target in node.targets:
            if isinstance(target, ast.Name):
                self._redefined.add(target.id)

        return self.generic_visit(node)

    def visit_FunctionDef(self, node):
        original_redefined = self._redefined.copy()

        self.generic_visit(node)

        self._redefined = original_redefined

        return node

    def visit_Name(self, node):
        if node.id in self._redefined:
            return node

        if node.id in self._aliases:
            return ast.Name(id=self._aliases[node.id], ctx=node.ctx)

        return node


class _ImportCollector(ast.NodeVisitor):
    def __init__(self):
        super().__init__()

        self.imports = []

    def visit_Import(self, node):
        self.imports.append(node)

        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        self.imports.append(node)

        self.generic_visit(node)


class _FunctionDefFinder(ast.NodeVisitor):
    def __init__(self, name):
        self._name = name

        self.result = None

    def visit_FunctionDef(self, node):
        if node.name == self._name:
            self.result = node

        self.generic_visit(node)



import ast
import inspect
import numbers
import types

import ninetoothed.naming as naming


class Symbol:
    def __init__(self, expr, constexpr=None, meta=None):
        if isinstance(expr, type(self)):
            self._node = expr._node
            return

        if isinstance(expr, ast.AST):
            self._node = expr
            return

        if isinstance(expr, types.CodeType):
            expr = inspect.getsource(expr)

        if not isinstance(expr, str):
            expr = str(expr)

        self._node = ast.parse(expr, mode="eval").body

        if (constexpr or meta) and not isinstance(self._node, ast.Name):
            raise ValueError("`constexpr` and `meta` are properties of name symbols.")

        if meta:
            if constexpr is False:
                raise ValueError("Non-constexpr meta symbol is not supported.")

            self._node.id = naming.make_meta(self._node.id)

        if constexpr:
            self._node.id = naming.make_constexpr(self._node.id)

    def __eq__(self, other):
        if isinstance(self._node, ast.Constant):
            if isinstance(other, Symbol) and isinstance(other._node, ast.Constant):
                return self._node.value == other._node.value

            if isinstance(other, numbers.Number):
                return self._node.value == other

        return False

    def __hash__(self):
        return id(self)

    def __add__(self, other):
        other = type(self)(other)

        if self == 0:
            return other

        if other == 0:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Add(), right=other._node))

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        other = type(self)(other)

        if self == 0:
            return -other

        if other == 0:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Sub(), right=other._node))

    def __rsub__(self, other):
        return type(self)(other).__sub__(self)

    def __mul__(self, other):
        other = type(self)(other)

        if self == 0 or other == 0:
            return type(self)(0)

        if self == 1:
            return other

        if other == 1:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Mult(), right=other._node))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __floordiv__(self, other):
        other = type(self)(other)

        if isinstance(other._node, ast.Constant) and other._node.value == 1:
            return self

        return type(self)(
            ast.BinOp(left=self._node, op=ast.FloorDiv(), right=other._node)
        )

    def __mod__(self, other):
        other = type(self)(other)

        return type(self)(ast.BinOp(left=self._node, op=ast.Mod(), right=other._node))

    def __lt__(self, other):
        other = type(self)(other)

        return type(self)(
            ast.Compare(left=self._node, ops=[ast.Lt()], comparators=[other._node])
        )

    def __and__(self, other):
        other = type(self)(other)

        return type(self)(
            ast.BinOp(left=self._node, op=ast.BitAnd(), right=other._node)
        )

    def __rand__(self, other):
        return self.__and__(other)

    def __getitem__(self, key):
        return type(self)(ast.Subscript(value=self._node, slice=type(self)(key)._node))

    def __repr__(self):
        return ast.unparse(self._node)

    def find_and_replace(self, target, replacement):
        _FindAndReplacer(target.node, replacement.node).visit(self._node)

    def names(self):
        class NameCollector(ast.NodeVisitor):
            def __init__(self):
                self.names = set()

            def visit_Name(self, node):
                self.generic_visit(node)

                self.names.add(node.id)

        name_collector = NameCollector()

        name_collector.visit(self._node)

        return name_collector.names

    @property
    def node(self):
        class SliceSimplifier(ast.NodeTransformer):
            def visit_Call(self, node):
                if isinstance(node.func, ast.Name) and node.func.id == "slice":
                    return ast.Slice(*node.args)

                return node

        return SliceSimplifier().visit(self._node)

    @staticmethod
    def is_name(object):
        return isinstance(object, Symbol) and isinstance(object.node, ast.Name)


class _FindAndReplacer(ast.NodeTransformer):
    def __init__(self, target, replacement):
        self._target_id = target.id
        self._replacement = replacement

    def visit_Name(self, node):
        if node.id == self._target_id:
            return self._replacement

        return self.generic_visit(node)
